/*
 * Copyright (c) 2015, 2015, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.nodes.qirinterface;

import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import com.oracle.truffle.api.frame.Frame;
import com.oracle.truffle.api.frame.FrameDescriptor;
import com.oracle.truffle.api.frame.FrameSlot;
import com.oracle.truffle.api.source.Source;
import com.oracle.truffle.api.source.SourceSection;
import com.oracle.truffle.r.nodes.function.FunctionDefinitionNode;
import com.oracle.truffle.r.nodes.function.FunctionExpressionNode;
import com.oracle.truffle.r.runtime.RRuntime;
import com.oracle.truffle.r.runtime.context.RContext;
import com.oracle.truffle.r.runtime.data.RFunction;
import com.oracle.truffle.r.runtime.data.RPromise;
import com.oracle.truffle.r.runtime.env.REnvironment;

import qir.ast.*;
import qir.ast.data.*;
import qir.ast.expression.*;
import qir.driver.sql.OracleDriver;
import qir.driver.sql.PostgreSQLDriver;
import qir.driver.ConnectionData;
import qir.driver.DBDriver;
import qir.driver.QIRDriver;
import qir.driver.hbase.HBaseDriver;

public final class QIRInterface {
    private static final Map<String, DBDriver<?, ?>> driverCache = new HashMap<>();

    /**
     * Runs a query in the database pointed by the given driver and returns the results.
     *
     * @param query The query to be executed in the targeted database
     * @return The results of the query
     */
    public static final Map<String, List<Object>> run(final QIRNode query) {
        // Run the query and retrieve results
        QIRNode qirRes = QIRDriver.run(query);

        // An error during the evaluation of the query
        if (qirRes == null)
            throw new RuntimeException("Could not run query: " + query);

        // The query did not return the right type
        if (!(qirRes instanceof QIRList))
            throw new RuntimeException("Internal error: query returned a non-list type");

        // The query returned an empty result
        if (qirRes == QIRLnil.instance)
            return new HashMap<>();

        final Map<String, List<Object>> results = new HashMap<>();

        // Inhabit the result columns with empty lists (we need the keys to create the lists) TODO:
        // Find a way to add the first row here and skip it on the next loop
        for (QIRNode firstRow = ((QIRLcons) qirRes).value; firstRow != QIRTnil.instance; firstRow = ((QIRTcons) firstRow).tail)
            results.put(((QIRTcons) firstRow).id, new ArrayList<>());

        // Process the rest of the rows TODO: Error if a row has unknown column
        for (; qirRes != QIRLnil.instance; qirRes = ((QIRLcons) qirRes).tail)
            for (QIRNode row = ((QIRLcons) qirRes).value; row != QIRTnil.instance; row = ((QIRTcons) row).tail)
                results.get(((QIRTcons) row).id).add(QIRToRType(((QIRTcons) row).value));

        return results;
    }

    /**
     * Resolves the free variables of a query.
     *
     * @param arg The query to be executed in the targeted database
     * @param argFrame The environment of execution
     * @return The closed query
     */
    public static final QIRNode normalize(final QIRNode arg, final Frame argFrame) {
        QIRNode query = arg;
        final SourceSection dummy = Source.newBuilder("").name("QIR interface").mimeType(RRuntime.R_APP_MIME).build().createUnavailableSection();

        for (Map<String, QIRVariable> fvs = QIRDriver.getFreeVars(query); !fvs.isEmpty(); fvs = QIRDriver.getFreeVars(query))
            for (final QIRVariable fv : fvs.values()) {
                final String varName = fv.id;
                Frame frame = argFrame;
                FrameSlot varSlot = frame.getFrameDescriptor().findFrameSlot(varName);
                for (; varSlot == null && frame.getArguments()[4] instanceof Frame && ((Frame) frame.getArguments()[4]).getArguments()[4] instanceof Frame; frame = (Frame) frame.getArguments()[4])
                    varSlot = ((Frame) frame.getArguments()[4]).getFrameDescriptor().findFrameSlot(varName);
                query = new QIRApply(dummy, new QIRLambda(dummy, null, new QIRVariable(dummy, varName, varSlot), query, new FrameDescriptor()),
                                varName.equals("interval") ? new QIRBuiltin(dummy, "interval") : varName.equals("having") ? new QIRBuiltin(dummy, "having") // TODO:
                                                                                                                                                            // Remove
                                                                                                                                                            // this
                                                                                                                                                            // hack
                                                : RToQIRType(fv.sourceSection, varSlot != null ? frame.getValue(varSlot) : RContext.lookupBuiltin(varName)));
            }
        return query;
    }

    /**
     * Returns a {@link DBDriver} of a database using a configuration file. This function uses a
     * cache for the drivers.
     *
     * @param dbName The name of the targeted database
     * @param fileName The name of the configuration file
     * @return A {@link DBDriver} to the given database with the given configuration
     */
    public static final DBDriver<?, ?> createDriver(final String dbName, final String fileName) throws UnsupportedOperationException {
        final String cacheKey = dbName + '@' + fileName;
        DBDriver<?, ?> result = driverCache.get(cacheKey);

        if (result != null)
            return result;
        switch (dbName) {
            case "Oracle":
                result = new OracleDriver(createConnectionData("ora", fileName));
                break;
            case "PostgreSQL":
                result = new PostgreSQLDriver(createConnectionData("pgsql", fileName));
                break;
            case "HBase":
                result = new HBaseDriver(fileName);
                break;
            default:
                throw new RuntimeException("Unknown database: " + dbName);
        }
        driverCache.put(cacheKey, result);
        return result;
    }

    /**
     * This function closes the connection of drivers created with the function
     * {@link #createDriver(String, String)}.
     */
    public static final void closeDrivers() {
        driverCache.values().stream().forEach(driver -> driver.closeConnection());
    }

    /**
     * Creates a {@link ConnectionData} object based on the given configuration file.
     *
     * @param prefix A prefix for properties in the configuration file
     * @param fileName The name of the configuration file
     * @return A {@link ConnectionData} created from the given configuration file
     */
    private static final ConnectionData createConnectionData(final String prefix, final String fileName) {
        try {
            final Reader dbConfigReader = new FileReader(fileName);
            final Properties p = new Properties();
            p.load(dbConfigReader);
            final String sessionId = p.getProperty(prefix + ".sid");
            final String serverName = p.getProperty(prefix + ".host");
            final int port = Integer.parseInt(p.getProperty(prefix + ".port"));
            final String userName = p.getProperty(prefix + ".user");
            final String passwd = p.getProperty(prefix + ".passwd");
            return new ConnectionData(sessionId, serverName, port, userName, passwd);
        } catch (IOException e) {
            throw new RuntimeException("Invalid configuration file for connection to a database: " + fileName);
        }
    }

    /**
     * Translates a QIR expression into a R statement.
     *
     * @param value The QIR expression to translate
     * @return The translation of the QIR expression in R
     * @throws UnsupportedOperationException If the type of the value is not supported.
     */
    @SuppressWarnings("unchecked")
    private static final <T> T QIRToRType(final QIRNode value) throws UnsupportedOperationException {
        if (value instanceof QIRBaseValue<?>)
            return ((QIRBaseValue<T>) value).value;
        throw new RuntimeException("Unsupported value: " + value);
    }

    /**
     * Translates a R statement into a QIR expression.
     *
     * @param src The {@link SourceSection} of the given value
     * @param value The R statement to translate
     * @return The translation of the R statement in QIR
     */
    static final <T> QIRNode RToQIRType(final SourceSection src, final T value) {
        if (value instanceof BigInteger)
            return new QIRBigNumber(src, (BigInteger) value);
        if (value instanceof Long)
            return new QIRNumber(src, (Long) value);
        if (value instanceof Byte)
            return new QIRNumber(src, (Byte) value);
        if (value instanceof Double)
            return new QIRDouble(src, (Double) value);
        if (value instanceof Boolean)
            return QIRBoolean.create((Boolean) value);
        if (value instanceof String)
            return new QIRString(src, (String) value);
        if (value instanceof REnvironment) {
            REnvironment env = (REnvironment) value;
            final Object queryId = env.get("queryId");
            if (queryId != null) // The object is a query
                return normalize(RContext.queries.get((Integer) queryId), RContext.envs.get((Integer) queryId));
            final String tableName = (String) env.get("tableName");
            final String schemaName = (String) env.get("schemaName");
            final String dbName = (String) env.get("dbName");
            final String configFile = (String) env.get("configFile");
            if (tableName != null && dbName != null && configFile != null) // The object is a table
                return new QIRTable<>(src, new QIRString(null, tableName), new QIRString(null, schemaName), createDriver(dbName, configFile));
            // else, the object is considered a tuple
            QIRTuple tuple = QIRTnil.instance;
            for (final Object x : env.getFrame().getFrameDescriptor().getIdentifiers())
                tuple = new QIRTcons(src, (String) x, RToQIRType(src, env.get((String) x)), tuple);
            return tuple;
        }
        if (value instanceof FunctionExpressionNode) {
            return RFunctionToQIRType(src, ((FunctionExpressionNode) value).getSyntaxDebugName(), (FunctionDefinitionNode) (((FunctionExpressionNode) value).getCallTarget().getRootNode()));
        }
        if (value instanceof RFunction) {
            final RFunction fun = (RFunction) value;
            switch (fun.getName()) {
                case "new.env":
                    return new QIRLambda(src, "new.env", new QIRVariable(null, "_", null), QIRTnil.instance, new FrameDescriptor());
                case "return":
                case "(":
                case "query.force": {
                    final QIRVariable x = new QIRVariable(null, "x", null);
                    return new QIRLambda(src, "identity", x, x, new FrameDescriptor());
                }
                case "new.tableRef":
                    final QIRVariable tableName = new QIRVariable(null, "__tmp__");
                    final QIRVariable schemaName = new QIRVariable(null, "__tmp2__");
                    final QIRVariable dummyVar = new QIRVariable(null, "__dummy__");
                    return new QIRLambda(null, null, tableName,
                                    new QIRLambda(null, null, dummyVar, new QIRLambda(null, null, dummyVar,
                                                    new QIRLambda(null, null, schemaName, new QIRTable<>(null, tableName, schemaName, null), new FrameDescriptor()), new FrameDescriptor()),
                                                    new FrameDescriptor()),
                                    new FrameDescriptor());
                case "sum":
                case "min":
                case "max":
                case "date":
                    return new QIRBuiltin(src, fun.getName());
                case "substr":
                    return new QIRBuiltin(src, "substring");
                case "mean":
                    return new QIRBuiltin(src, "avg");
                case "length":
                    return new QIRBuiltin(src, "count");
                case "match":
                    return new QIRBuiltin(src, "match");
                case "format": {
                    final QIRVariable x = new QIRVariable(null, "x", null);
                    final QIRVariable y = new QIRVariable(null, "y", null);
                    return new QIRLambda(null, null, x, new QIRLambda(null, null, y, new QIRApply(null, new QIRBuiltin(src, "extractYear"), x), new FrameDescriptor()),
                                    new FrameDescriptor());
                }
                case "regexpr":
                    final QIRVariable x = new QIRVariable(null, "x", null);
                    final QIRVariable y = new QIRVariable(null, "y", null);
                    return new QIRLambda(null, null, x, new QIRLambda(null, null, y, new QIRApply(null, new QIRApply(null, new QIRBuiltin(src, "like"), y), x), new FrameDescriptor()),
                                    new FrameDescriptor());
                default:
                    return RFunctionToQIRType(src, fun.getName(), (FunctionDefinitionNode) fun.getRootNode());
            }
        }
        if (value instanceof RPromise) {
            final RPromise fun = (RPromise) value;
            if (fun.isEvaluated())
                return RToQIRType(src, fun.getValue());
            return RToQIRType(src, fun.getClosure().eval(fun.getFrame()));
        }
        throw new RuntimeException("Unsupported value: " + value);
    }

    private static final QIRNode RFunctionToQIRType(final SourceSection src, final String funName, final FunctionDefinitionNode fun) {
        try {
            QIRNode res = ((FunctionDefinitionNode) fun.getCallTarget().getRootNode()).getBody().accept(QIRTranslateVisitor.instance);
            final String[] args = ((FunctionDefinitionNode) fun.getCallTarget().getRootNode()).getSignature().getNames();

            if (args.length == 0)
                return new QIRLambda(src, funName, null, res, new FrameDescriptor());
            for (int i = args.length - 1; i >= 0; i--)
                res = new QIRLambda(src, funName, new QIRVariable(null, args[i], null), res, new FrameDescriptor());
            return res;
        } catch (UnsupportedOperationException e) {
            final SourceSection funSrc = fun.getCallTarget().getRootNode().getSourceSection();
            // TODO: Handle dependencies
            return new QIRTruffleNode(funSrc, "R", funSrc.getCode());
        }
    }
}